/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "utils.h"
#include "shared.h"

/*------------------------------------------------------------------*/

/*  Load_Line()
 *
 *  Loads a line from a file, aborts on failure. Lines beginning
 *  with a '#' are ignored as comments. At the end of file EOF is
 *  returned. Lines assumed maximum 195 characters long.
 */

  int
Load_Line( char *buff, FILE *pfile, char *mesg )
{
  int
    num_chr, /* Number of characters read, excluding lf/cr */
    chr;     /* Character read by getc() */
  char error_mesg[MESG_SIZE];

  /* Prepare error message */
  snprintf( error_mesg, sizeof(error_mesg),
      _("Error reading %s\n"\
        "Premature EOF (End Of File)"), mesg );

  /* Clear buffer at start */
  buff[0] = '\0';
  num_chr = 0;

  /* Get next character, return error if chr = EOF */
  if( (chr = fgetc(pfile)) == EOF )
  {
    fprintf( stderr, "xgridloc: %s\n", error_mesg );
    fclose( pfile );
    Error_Dialog( error_mesg );
    return( EOF );
  }

  /* Ignore commented lines and eol/cr */
  while(
      (chr == '#') ||
      (chr == ' ') ||
      (chr == HT ) ||
      (chr == CR ) ||
      (chr == LF ) )
  {
    /* Go to the end of line (look for LF or CR) */
    while( (chr != CR) && (chr != LF) )
      /* Get next character, return error if chr = EOF */
      if( (chr = fgetc(pfile)) == EOF )
      {
        fprintf( stderr, "xgridloc: %s\n", error_mesg );
        fclose( pfile );
        Error_Dialog( error_mesg );
        return( EOF );
      }

    /* Dump any CR/LF remaining */
    while( (chr == CR) || (chr == LF) )
      /* Get next character, return error if chr = EOF */
      if( (chr = fgetc(pfile)) == EOF )
      {
        fprintf( stderr, "xgridloc: %s\n", error_mesg );
        fclose( pfile );
        Error_Dialog( error_mesg );
        return( EOF );
      }

  } /* End of while( (chr == '#') || ... */

  /* Continue reading characters from file till
   * number of characters = 195 or EOF or CR/LF */
  while( num_chr < 195 )
  {
    /* If LF/CR reached before filling buffer, return line */
    if( (chr == LF) || (chr == CR) ) break;

    /* Enter new character to line buffer */
    buff[num_chr++] = (char)chr;

    /* Get next character */
    if( (chr = fgetc(pfile)) == EOF )
    {
      /* Terminate buffer as a string if chr = EOF */
      buff[num_chr] = '\0';
      return( SUCCESS );
    }

    /* Abort if end of line not reached at 195 char. */
    if( (num_chr == 195) && (chr != LF) && (chr != CR) )
    {
      /* Terminate buffer as a string */
      buff[num_chr] = '\0';
      snprintf( error_mesg, sizeof(error_mesg),
          _("Error reading %s\n"\
            "Line longer than 195 characters"), mesg );
      fprintf( stderr, "xgridloc: %s\n%s\n", error_mesg, buff );
      fclose( pfile );
      Error_Dialog( error_mesg );
      return( ERROR );
    }

  } /* End of while( num_chr < max_chr ) */

  /* Terminate buffer as a string */
  buff[num_chr] = '\0';

  return( SUCCESS );

} /* End of Load_Line() */

/*------------------------------------------------------------------*/

/*  Load_Config()
 *
 *  Loads the xgridloc configuration file
 */

  gboolean
Load_Config( gpointer data )
{
  char
    dist_unit[10],  /* Name of distance unit   */
    *rc_fpath,      /* File path to xgridlocrc */
    line[196];      /* Buffer for Load_Line    */

  int idx, len;

  /* Config file pointer */
  FILE *xgridlocrc;


  /* Setup file path to xgridlocrc */
  size_t s = strlen(getenv("HOME")) + strlen(".xgridloc/xgridlocrc") + 2;
  rc_fpath = (char *)malloc( s );
  if( rc_fpath == NULL )
  {
    Error_Dialog(
        _("Failed to allocate memory for\n"\
          "Configuration file name string.\n"\
          "Please quit xgridloc and correct") );
    return( FALSE );
  }
  snprintf( rc_fpath, s, "%s/%s", getenv("HOME"), ".xgridloc/xgridlocrc" );
  rc_fpath[s-1] = '\0';

  /* Open ~./xgridlocrc file */
  xgridlocrc = fopen( rc_fpath, "r" );
  if( xgridlocrc == NULL )
  {
    Error_Dialog(
        _("Error opening configuration file\n"\
          "Please quit xgridloc and correct") );
    fprintf( stderr, "Xgridloc: error opening configuration file: %s\n", rc_fpath );
    free(rc_fpath);
    return( FALSE );
  }

  free(rc_fpath);

  /*** Read runtime configuration data ***/
  /* Read 'Home' position, abort on EOF */
  if( Load_Line(line, xgridlocrc, _("Home Position") ) != SUCCESS )
    return( FALSE );
  Strlcpy( rc_data.home_posn, line, sizeof(rc_data.home_posn) );

  /* Read 'Home' position name, abort if EOF */
  if( Load_Line(line, xgridlocrc, _("Home Position Name") ) != SUCCESS )
    return( FALSE );
  Strlcpy( rc_data.home_name, line, sizeof(rc_data.home_name) );

  /* Read distance unit name and conversion factor, abort if EOF */
  if( Load_Line(line, xgridlocrc, _("Distance Unit Name") ) != SUCCESS )
    return( FALSE );
  len = (int)strlen( line );
  for( idx = 0; idx < len; idx++ )
    if( (line[idx] == ' ') || (idx == 9) )
      break;
  Strlcpy( dist_unit, line, (size_t)(idx + 1) );
  rc_data.unit_conv = Atof( &line[idx] );

  /* Set distance labels */
  gtk_label_set_text( GTK_LABEL(
        Builder_Get_Object(main_window_builder, "short_label")), dist_unit );
  gtk_label_set_text( GTK_LABEL(
        Builder_Get_Object(main_window_builder, "long_label")), dist_unit );

  /* Read xplanet command */
  if( Load_Line(rc_data.xplanet_cmnd, xgridlocrc, _("Xplanet Command") ) != SUCCESS )
    return( FALSE );

  /* Read xplanet config file */
  if( Load_Line(rc_data.config_file, xgridlocrc, _("Xplanet config file") ) != SUCCESS )
    return( FALSE );

  /* Read xplanet marker file */
  if( Load_Line(rc_data.marker_file, xgridlocrc, _("Xplanet marker file") ) != SUCCESS )
    return( FALSE );

  /* Read xplanet arc file */
  if( Load_Line(rc_data.arc_file, xgridlocrc, _("Xplanet arc file") ) != SUCCESS )
    return( FALSE );

  fclose( xgridlocrc );

  return( FALSE );

} /* End of Load_Config */

/*------------------------------------------------------------------*/

/*  Cleanup()
 *
 *  Cleans up before quitting
 */

int
Cleanup(void)
{
  /* Delete xplanet files */
  if( isFlagSet(XPLANET_ENABLED) )
  {
    unlink( rc_data.config_file );
    unlink( rc_data.marker_file );
    unlink( rc_data.arc_file );
    return( system("pkill xplanet") );
  }
  return( 0 );
}

/*------------------------------------------------------------------------*/

/* Functions for testing and setting/clearing flags */

/* An int variable holding the single-bit flags */
static int Flags = 0;

int
isFlagSet(int flag)
{
  return(Flags & flag);
}

int
isFlagClear(int flag)
{
  return(~Flags & flag);
}

void
SetFlag(int flag)
{
  Flags |= flag;
}

void
ClearFlag(int flag)
{
  Flags &= ~flag;
}

void
ToggleFlag(int flag)
{
  Flags ^= flag;
}

/*------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage( void )
{
  fprintf( stderr, "%s\n",
      _("Usage: xgridloc [-hv]") );

  fprintf( stderr, "%s\n",
      _("       -h: Print this usage information and exit"));

  fprintf( stderr, "%s\n",
      _("       -v: Print version number and exit"));

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/* Strlcpy()
 *
 * Copies n-1 chars from src string into dest string. Unlike other
 * such library fuctions, this makes sure that the dest string is
 * null terminated by copying only n-1 chars to leave room for the
 * terminating char. n would normally be the sizeof(dest) string but
 * copying will not go beyond the terminating null of src string
 */
  void
Strlcpy( char *dest, const char *src, size_t n )
{
  char ch = src[0];
  int idx = 0;

  /* Leave room for terminating null in dest */
  n--;

  /* Copy till terminating null of src or to n-1 */
  while( (ch != '\0') && (n > 0) )
  {
    dest[idx] = src[idx];
    idx++;
    ch = src[idx];
    n--;
  }

  /* Terminate dest string */
  dest[idx] = '\0';

} /* Strlcpy() */

/*------------------------------------------------------------------*/

/* Strlcat()
 *
 * Concatenates at most n-1 chars from src string into dest string.
 * Unlike other such library fuctions, this makes sure that the dest
 * string is null terminated by copying only n-1 chars to leave room
 * for the terminating char. n would normally be the sizeof(dest)
 * string but copying will not go beyond the terminating null of src
 */
  void
Strlcat( char *dest, const char *src, size_t n )
{
  char ch = dest[0];
  int idd = 0; /* dest index */
  int ids = 0; /* src  index */

  /* Find terminating null of dest */
  while( (n > 0) && (ch != '\0') )
  {
    idd++;
    ch = dest[idd];
    n--; /* Count remaining char's in dest */
  }

  /* Copy n-1 chars to leave room for terminating null */
  n--;
  ch = src[ids];
  while( (n > 0) && (ch != '\0') )
  {
    dest[idd] = src[ids];
    ids++;
    ch = src[ids];
    idd++;
    n--;
  }

  /* Terminate dest string */
  dest[idd] = '\0';

} /* Strlcat() */

/*------------------------------------------------------------------*/

/* Atof()
 *
 * Replaces atof() to take into account the
 * locale-dependent decimal point character
 */
double Atof( const char *nptr )
{
  int idx;
  size_t len;
  double d = 0.0;
  char *s = NULL;
  static gboolean first_call = TRUE;
  static char dp = '.';

  /* Find locale-dependent decimal point character */
  if( first_call )
  {
    struct lconv *lcnv;
    lcnv = localeconv();
    dp = *lcnv->decimal_point;
    first_call = FALSE;
  }

  /* Look for a . or , decimal point character
   * in the supplied number buffer (string) */
  len = strlen( nptr );
  for( idx = 0; idx < (int)len; idx++ )
    if( (nptr[idx] == ',') || (nptr[idx] == '.') )
      break;

  /* Create temporary string to modify decimal point */
  s = malloc( len + 1 );
  if( s == NULL )
  {
    Error_Dialog(
        _("Failed to allocate memory for\n"\
          "Atof() temporary string buffer.\n"\
          "Please quit xgridloc and correct") );
    return( d );
  }
  strcpy( s, nptr );
  s[ len ] = '\0';

  /* If a decimal point character is found, replace */
  if( idx < (int)len ) s[idx] = dp;
  d = atof( s );
  free( s );

  return( d );
} /* End of Atof() */

/*------------------------------------------------------------------*/
